__all__ = ()

APPLICATION_COMMAND_OPTION_NAME_LENGTH_MIN = 1
APPLICATION_COMMAND_OPTION_NAME_LENGTH_MAX = 32
APPLICATION_COMMAND_OPTION_DESCRIPTION_LENGTH_MIN = 2
APPLICATION_COMMAND_OPTION_DESCRIPTION_LENGTH_MAX = 100
