__all__ = ()

APPLICATION_COMMAND_OPTION_OPTIONS_MAX = 25
APPLICATION_COMMAND_OPTION_CHOICES_MAX = 25

APPLICATION_COMMAND_OPTION_MAX_LENGTH_DEFAULT = 0
APPLICATION_COMMAND_OPTION_MAX_LENGTH_MIN = 1
APPLICATION_COMMAND_OPTION_MAX_LENGTH_MAX = 6000

APPLICATION_COMMAND_OPTION_MIN_LENGTH_DEFAULT = 0
APPLICATION_COMMAND_OPTION_MIN_LENGTH_MIN = 0
APPLICATION_COMMAND_OPTION_MIN_LENGTH_MAX = 6000
