__all__ = ()

BOT_PUBLIC_DEFAULT = False
BOT_REQUIRE_CODE_GRANT_DEFAULT = False
OVERLAY_DEFAULT = False
OVERLAY_COMPATIBILITY_HOOK_DEFAULT = False
HOOK_DEFAULT = False


MAX_PARTICIPANTS_DEFAULT = 0

NAME_LENGTH_MIN = 2
NAME_LENGTH_MAX = 128

DESCRIPTION_LENGTH_MAX = 400
